/*************************************************************************/
/*************************************************************************/
/*                                                                       */
/*  File Name : VRPSolution.cc                                           */
/*                                                                       */
/*       Date : May  5, 1999                                             */
/* Updated at : May  6, 1999                                             */
/*                                                                       */
/*     Author : H. Timucin OZDEMIR                                       */
/*                                                                       */
/*************************************************************************/
/*************************************************************************/

#include <iostream.h>
#include <iomanip.h>
#include <stdio.h>
#include <stdlib.h>

#include <RandomGenerator.h>

#include "VRPcommons.h"
#include "VRProblem.h"
#include "VRPSolution.h"
#include "Customer.h"
#include "CustomerList.h"
#include "Depot.h"
#include "DepotList.h"
#include "EdgeGraph.h"
#include "ListOfEdgeLists.h"
#include "EdgeList.h"
#include "VRPUtil.h"

/*************************************************************************/
/*************************************************************************/
VRPSolution::VRPSolution(VRProblem* _myProblem) {
  myProblem = _myProblem;
  wtime=tdemand=cost = 0;
  myCover = new Cover(myProblem->getNumberOfCustomers());
}// end of constructor

/*************************************************************************/
/*************************************************************************/
VRPSolution::VRPSolution() {
  myProblem = NULL;
  wtime=tdemand=cost = 0;
  myCover = NULL;
}// end of constructor

/*************************************************************************/
/*************************************************************************/
VRPSolution::~VRPSolution() { dispose(); }

/*************************************************************************/
// copy constructor
/*************************************************************************/
VRPSolution::VRPSolution(const VRPSolution& rhs) {  copyFrom(rhs); }

/*************************************************************************/
/*************************************************************************/
void VRPSolution::setProblem(VRProblem* _myProblem) 
{ myProblem=_myProblem; 
  myCover = new Cover(myProblem->getNumberOfCustomers()); 
  wtime=tdemand=cost = 0;
  cleanUpVehicles();
}// end of setProblem()

/*************************************************************************/
/*************************************************************************/
void VRPSolution::addVehicleSchedule(ListOfEdgeLists& eList) 
{  int i;
   EdgeList *el;
   //cout << " in addVehicleSchedule " << eList << endl;

   for(i=0;i<eList.getNumberOfEdgeLists();i++)
   { if((el=eList.getNthEdgeList(i))!=NULL) addVehicleDuty(el); }

}// end of addVehicleSchedule()

/*************************************************************************/
/** If it can merge, then it does nothing. But it does not, then it 
    defines another vehcile schedule and adds this edge list to the
    vehicle's schedule.

//  methods : preProcess()
//            canMerge()
//            myEdgeLists.addElement()
//
**/
/*************************************************************************/
void VRPSolution::addVehicleDuty(EdgeListPtr edgeList) 
{  
   EdgeListPtr nList = preProcess(edgeList);
   if(nList!=NULL&&nList->getNumberOfEdges()>1&&!canMerge(nList))
   {
      ListOfEdgeListsPtr nListOfEdgeList = new ListOfEdgeLists(nList);
      myEdgeLists.addElement(nListOfEdgeList);

      // Update Cover
      myCover->updateCover(nListOfEdgeList);

      // Add the cost of this Schedule
      cost += nListOfEdgeList->getDCost();

      // Add the demand of this Schedule
      tdemand += nListOfEdgeList->getDemand();

      // Add the waiting time of this Schedule
      wtime += nListOfEdgeList->getWaitingTime();

   }// end of if()

}// end of addVehicleDuty

/*************************************************************************/
/** Let's remove possible overcoverings */
/*************************************************************************/
/*
EdgeListPtr VRPSolution::processEList(EdgeListPtr edgeList) {

   int i,notRequired=1;
   for(i=1;notRequired&&i<getNumberOfEdges();i++) 
     if( (curr = edgeList->getNthEdge(i))!=NULL )
       notRequired=!myCover->isCovered(curr->getStartNode());
   EdgeListPtr ret = edgeList;
   if(!notRequired) {
     ret = new EdgeList();
     int currNode = edgeList->getStartNode();
     vrptime startD = edgeList->getStartDate();
     for(i=1;notRequired&&i<getNumberOfEdges();i++) 
        if( (curr = edgeList->getNthEdge(i))!=NULL && 
	    !myCover->isCovered(curr->getStartNode()) ) {
	    Edge aEdge(myProblem,currNode,curr->getStartNode(),startD);
	    ret->addEdge(aEdge);
	    currNode=aEdge.getEndNode();
	    startD=aEdge.getStartDateSecondNode();
        }// end of if
     if(currNode!=edgeList->getStartNode()) {
       Edge aEdge(myProblem,currNode,edgeList->getStartNode(),startD);
       ret->addEdge(aEdge);
     }// end of if
     if(ret->getNumberOfEdges()<2) { delete ret; ret=NULL; }
   }//
   return ret;
}// end of processEList()
*/

/*************************************************************************/
/** Is it possible for this EdgeList to merge with current vehicle
    schedules ?
//
// methods : getNumberOfVehicles()
//           ListOfEdgeLists::canMerge()
//           ListOfEdgeLists::addEdgeList()
**/
/*************************************************************************/
VRPSolution::canMerge(EdgeListPtr edgeList)
{ int ret = 0;
  int i=0;
  ListOfEdgeListsPtr tmp;
  int minIdx = -1;
  double tofFT = FLT_MAX;

  // It prefers the minimum cost schedule
  while( i<getNumberOfVehicles() )
  {
     tmp = getListOfEdgeListsIndex(i);
     if( tmp!=NULL && tmp->canMerge(edgeList,myProblem) && 
		       tmp->getDCost()<tofFT )
     { minIdx = i; tofFT = tmp->getDCost(); }
     i++;

  }// end of while

  // if there is such vehicle schedule
  if( minIdx>-1 )
  {
     tmp = getListOfEdgeListsIndex(minIdx);
     if( tmp!=NULL) 
     {
	// Remove the cost of this schedule
	cost -= tmp->getDCost();
	tdemand -= tmp->getDemand();
	wtime -= tmp->getWaitingTime();

        tmp->addEdgeList(edgeList,myProblem);

	// Add the cost of this schedule
	cost += tmp->getDCost();
	tdemand += tmp->getDemand();
	wtime += tmp->getWaitingTime();

        ret = 1;

	// Update the cover
        myCover->updateCover(edgeList);

      }// end of if

   }// end of if

   return ret;

}// end of canMerge()

/*************************************************************************/
/** Resets the schedule list. */
/*************************************************************************/
void VRPSolution::cleanUpVehicles() {  
   dispose(); 
   if(myProblem!=NULL) myCover = new Cover(myProblem->getNumberOfCustomers()); 
   wtime=cost=tdemand=0;
}// end of cleanUpVehicles()

/*************************************************************************/
// assignment operator
/*************************************************************************/
VRPSolution& VRPSolution::operator=(const VRPSolution& rhs)
{  dispose();
   copyFrom(rhs);
   return *this;
}// end of operator=()

/*************************************************************************/
// equal?
/*************************************************************************/
VRPSolution::operator==(const VRPSolution& rhs)
{  int retVal=0;
   return retVal;
}// end of operator==()

/*************************************************************************/
/*************************************************************************/
void VRPSolution::dispose() { 
  ListOfEdgeListsPtr tmp;
  int i=0;
  for(i=0;i<getNumberOfVehicles();i++)
     if((tmp = getListOfEdgeListsIndex(i))!=NULL) delete tmp;
  myEdgeLists.reset(); 
  if(myCover!=NULL) { delete myCover; myCover=NULL; }
}

/*************************************************************************/
/*************************************************************************/
void VRPSolution::copyFrom(const VRPSolution& rhs) 
{  cost = rhs.cost;
   tdemand = rhs.tdemand;
   myProblem = NULL;
   myCover = NULL;
   if(rhs.myProblem!=NULL) {
      myProblem = rhs.myProblem;
      cleanUpVehicles(); 
      ListOfEdgeLists* tmp;
      for(int i=0;i<rhs.getNumberOfVehicles();i++)
       if((tmp=rhs.getListOfEdgeListsIndex(i))!=NULL) addVehicleSchedule(*tmp);
   }// end of if
}// end of copyFrom()

/*************************************************************************/
// print 
/*************************************************************************/
void VRPSolution::nicePrint(ostream& os) const {

   ListOfEdgeListsPtr tmp;
   int i=0;
   vrpcost tcost=0;

   for(i=0;i<getNumberOfVehicles();i++)
     if((tmp = getListOfEdgeListsIndex(i))!=NULL) 
     { os << setw(3) << (i+1) << " " << *tmp << endl; tcost+=tmp->getDCost(); }
   os <<" Number of Vehicles .: " << getNumberOfVehicles() << endl;
   os <<" Cost ...............: " << cost << endl;
   os <<" TCost ..............: " << tcost << endl;
   os <<" TDemand ............: " << tdemand << endl;
   os << *myCover << endl;

}// end of operator<<()

/*************************************************************************/
// print 
/*************************************************************************/
ostream& operator<<(ostream& os,const VRPSolution& rhs)
{
   rhs.nicePrint(os);
   return os;

}// end of operator<<()

/*************************************************************************/
// write number of total distance and number of vehicles  
/*************************************************************************/
void VRPSolution::writeStat(ostream& os) 
{
   os << setw(3) << myCover->getNumberOfUncovered() << " " 
      << setw(3) << getNumberOfVehicles() << " " 
      << cost << " ";

}// end of writeStat()

/*************************************************************************/
// write serialized schedule
/*************************************************************************/
void VRPSolution::writeCoded(ostream& os) {
  int i,j,k;
  ListOfEdgeListsPtr tmp;
  EdgeListPtr tmp2;
  EdgePtr     tmp3;
  Customer    cF;
  Customer    cT;
  os << getNumberOfVehicles() << endl;
  for(i=0;i<getNumberOfVehicles();i++)
    if((tmp = getListOfEdgeListsIndex(i))!=NULL) { 
      os << tmp->getNumberOfEdgeLists() << endl;
      for(j=0;j<tmp->getNumberOfEdgeLists();j++)
        if((tmp2 = tmp->getNthEdgeList(j))!=NULL) { 
           os << tmp2->getNumberOfEdges() << endl;
           for(k=0;k<tmp2->getNumberOfEdges();k++)
             if( (tmp3 = tmp2->getNthEdge(k))!=NULL&&
		  myProblem->getCustomer(tmp3->getStartNode(),cF)&&
		  myProblem->getCustomer(tmp3->getEndNode(),cT)) 
                  os << cF.getNodeNumber()   << " "
                     << cT.getNodeNumber()   << " " 
		     << tmp3->getStartDate() << endl;
        }// end of if
    }// end of if
}// end of writeCoded()

/*************************************************************************/
// read serialized schedule
/*************************************************************************/
void VRPSolution::readCoded(istream& is) {
  int i,j,k;
  ListOfEdgeListsPtr tmp;
  EdgeListPtr tmp2;
  Customer    cF;
  Customer    cT;
  int         nofv,nofel,nofe;
  customerId  from, to;
  vrptime     startt;
  EdgeGraph*  eGraph = myProblem->getEdgeGraph(); 
  cleanUpVehicles();
  is >> nofv;
  cout << " nofv : " << nofv << endl;
  for(i=0;i<nofv;i++) {
    tmp = new ListOfEdgeLists();
    is >> nofel;
    cout << " nofel : " << nofel << endl;
    for(j=0;j<nofel;j++) {
      tmp2 = new EdgeList();
      is >> nofe;
      cout << " nofe : " << nofe << endl;
      for(k=0;k<nofe;k++) {
	is >> from >> to >> startt;
	Edge tmp3( eGraph,myProblem->getNodeIndex(from),
		   myProblem->getNodeIndex(to),startt);
        tmp2->addEdge(tmp3); 
	cout << " added " << tmp3 << endl;
      }// end of for
      tmp->addEdgeList(tmp2,myProblem);
      cout << " added " << *tmp2 << endl;
    }// end of for
    addVehicleSchedule(*tmp);
    cout << " added " << *tmp << endl;
    delete tmp;
  }// end of for
}// end of readCoded()

/*************************************************************************/
// updates the vehicle schedules from ... 
// methods : addVehicleDuty
/*************************************************************************/
void VRPSolution::updateVehicles(ListOfEdgeListsPtr lofel) {
   //cout <<" updateVehicles: " << *lofel << endl;
   EdgeListPtr eList=NULL;
   int i=0;
   for(i=0;i<lofel->getNumberOfEdgeLists();i++)
     if((eList=lofel->getNthEdgeList(i))!=NULL) addVehicleDuty(eList);
}// end of updateVehicles()

/*************************************************************************/
// updates the vehicle schedules from ... 
//
// return a new EdgeList instance...
//
/*************************************************************************/
EdgeListPtr VRPSolution::preProcess(EdgeListPtr eList)
{
   //cout << " preProcess " << endl; 
   //eList->printTabular(cout);
   EdgeListPtr ret = new EdgeList();
   EdgePtr     crrEdge;
   EdgePtr     anEdge;
   vrptime     cTime=0;
   Customer    cus;
   int i=0;
   int* indx;
   gVector<int> collect(eList->getNumberOfEdges());
   if(eList->getNumberOfEdges()>0)
   {  EdgeGraphPtr eGraph = myProblem->getEdgeGraph();

      for(i=0;i<eList->getNumberOfEdges();i++)
       if( (crrEdge=eList->getNthEdge(i))!=NULL&&
           !myCover->isCovered(crrEdge->getStartNode()) ||
	   eGraph->isDepot(crrEdge->getStartNode()) ) 
	     collect.addElement(crrEdge->getStartNode());

      if( crrEdge!=NULL&& !myCover->isCovered(crrEdge->getEndNode()) ||
	  eGraph->isDepot(crrEdge->getEndNode()) ) 
	  collect.addElement(crrEdge->getEndNode());

      if(collect.getIndex())
      {
	int toIndex;
	int fromIndex = *collect.getNthElement(0);
	myProblem->getCustomer(fromIndex,cus);
	cTime=cus.getStartTime();
	for(i=1;i<collect.getIndex();i++)
	{
	   toIndex = *collect.getNthElement(i);
	   if(fromIndex!=toIndex && eGraph->doesEdgeExist(fromIndex,toIndex,cTime) )
	   {  anEdge = new Edge(eGraph,fromIndex,toIndex,cTime);
	      ret->addEdge(*anEdge);
	      fromIndex=toIndex;
	      cTime = anEdge->getStartDateSecondNode();
	      delete anEdge;
	   }// end of if

	}// end of if
      }// end of if

      if(ret->getNumberOfEdges()&&!ret->isValid(myProblem,1)) {
	cout << " VRPSolution.preProcess() The following EdgeList is not acceptable!" << endl;
	ret->printTabular(cout);
	exit(1);
      }// end of if

   }// end of if

   //cout << " after preProcess " << endl;
   //ret->printTabular(cout);
   return ret;

}// end of preProcess()

/*************************************************************************/
/**
    Selects the best edge from the possible edges based on performance
    function.
    Selects a random edge from the possible edges.
**/
/*************************************************************************/
Edge VRPSolution::pickAnEdge(WeightedRandomPickerSet<EdgePtr>* leaving, SelectStrategy strategy) {
  EdgePtr ret=NULL;
  if(leaving->getSize()==1) ret = *leaving->randomPick();
  else {
    vrpcost cst = FLT_MAX;
    EdgePtr    anEdge;
    int i;
    switch(strategy) {
      case GREEDY : 
           for(i=0;i<leaving->getSize();i++) 
             if( leaving->getElementAtSafe(i,anEdge)&&
		 anEdge->getCostPerDemand()<cst ) {
	       cst=anEdge->getCostPerDemand();
	       ret=anEdge;
             }// end of if
           break;
      case DISTGREEDY : 
           for(i=0;i<leaving->getSize();i++) 
             if( leaving->getElementAtSafe(i,anEdge) && 
	         anEdge->getCost()<cst ) {
	       cst=anEdge->getCost();
	       ret=anEdge;
             }// end of if
           break;
       dafault: ret = *leaving->randomPick(); break; 
    }// end of switch 
  }
  return (ret==NULL?**leaving->randomPick():*ret);
}// end of pickAnEdge()

/*************************************************************************/
/**
    Selects an edge with random strategy.
**/
/*************************************************************************/
Edge VRPSolution::pickAnEdgeWithRndSS(WeightedRandomPickerSet<EdgePtr>* leaving) {
  SelectStrategy ss = RANDOM;
  int ssindx = RandomGenerator::generateInt(0,5)%5;
  switch(ssindx) {
    case 1  : ss = GREEDY; break;
    case 2  : ss = DISTGREEDY; break;
    //case 3  : ss = WTIME; break;
    //case 4  : ss = WTPERDEMAND; break;
    default : ss = RANDOM; 
  }// end of switch
  return pickAnEdge(leaving,ss);
}// end of pickAnEdgeWithRndSS()

/*************************************************************************/
/**
    Tries to cover uncovered nodes in 'aCover' without using the edges in
    'abondoned' set.
**/
/*************************************************************************/
void VRPSolution::patchUpRandom(Cover aCover, tSet<EdgePtr>* abondoned)
{  int retVal=0;
   char enter;

   // Get List of Depots
   DepotListPtr depots = myProblem->getDepotList();
   unsigned   dindex ; 
   customerId depotId ; 
   unsigned   currIndex ; 
   unsigned   startIndex ; 
   vrptime    currTime ; 
   vrpcapacity   currCapacity ; 

   // Get the EdgeGraph
   EdgeGraph* myGraph = myProblem->getEdgeGraph();

   // Get the number of customers
   unsigned   numOfCustomers = myProblem->getNumberOfCustomers();

   // Define a cover
   WeightedRandomPickerSet<EdgePtr> leaving(numOfCustomers);
   if(depots==NULL) cout << " depots is NULL! "<< endl;
   //cout << " getNumberOfUncovered : " << aCover.getNumberOfUncovered() << endl;
   int notExit = 1;

   // There is depots, still uncovered nodes
   while( depots!=NULL && depots->getSize()>0 && 
	  aCover.getNumberOfUncovered() && notExit ) 
   {  
     // Get the customerId of the depot
     dindex = RandomGenerator::generateInt(0,depots->getSize());

     // Get the index of customerId depot
     depots->getDepot(dindex%depots->getSize(),depotId);
     startIndex = currIndex = myProblem->getNodeIndex(depotId);
     currTime = 0;

     // Get the maximum capacity of a vehicle
     currCapacity = myProblem->getVehicleCapacity();

     //cout << " Depot is " << currIndex << "["<<depotId<<"]" << endl;

     // Get the nodes leaving from this node
     leaving.clearAll(); 
     myGraph->getLeavingEdges( currIndex,leaving,&aCover,currTime,
			       currCapacity,abondoned);

     //VRPUtil::println(&leaving);
     //cout << " currIndex " << currIndex << endl;

     if( leaving.getSize() )
     {
       // Randomly pick an edge
       EdgePtr ePtr = *leaving.randomPick();
       //a = pickAnEdge(&leaving,GREEDY); pickAnEdgeWithRndSS(&leaving);
       //Edge a = *ePtr;
       Edge a = pickAnEdgeWithRndSS(&leaving);
       a.setStartDate(currTime);

       // define an edge list with this edge. 
       EdgeListPtr eList = new EdgeList(a);

       cout << " Edge " << a << endl;

       aCover.covered(currIndex);    // update cover
       currIndex = a.getEndNode();   // update the current node
       aCover.covered(currIndex);    // update cover
       currTime = a.getEndDate();    // update the current time
       //currTime = a.getStartDateSecondNode();    // update the current time
       Customer c;
       if( myProblem->getCustomer(currIndex,c) ) // get the Customer
       {
         currCapacity-=c.getDemand();  // update the capacity
         //cout << " Customer " << c 
	 //     << " capacity " << currCapacity 
	 //     << " time " << currTime << endl;
         int stay = 1;
         while( stay && currCapacity>0 )  // if there is some space
         {
           // Get the nodes leaving from this node
           leaving.clearAll();
           myGraph->getLeavingEdges( currIndex,leaving,&aCover,currTime,
				     currCapacity,abondoned);
           //VRPUtil::println(&leaving);
	   //cout << " currIndex " << currIndex << endl;
           if( leaving.getSize()==0 ) stay=0;
           if( stay )
           {
	     vrptime sstime = currTime = a.getStartDateSecondNode();
             // Randomly pick an edge
             //a = **leaving.randomPick();
             a = pickAnEdgeWithRndSS(&leaving);
             //a = pickAnEdge(&leaving,GREEDY); 
             ///////a.setStartDate(currTime); 
             a.setStartDate(sstime); 
	     currTime = a.getEndDate();  // update the current time
             eList->addEdge(a);          // add this edge to the edge list
             currIndex = a.getEndNode(); // update the current node
             aCover.covered(currIndex);  // update cover
             if( myProblem->getCustomer(currIndex,c) ) 
		 currCapacity-=c.getDemand();  // update the capacity

             //cout << " Edge " << a 
	     //     << " node " << currIndex 
	     //     << " capacity " << currCapacity 
	     //     << " time " << currTime << endl;
  
	     // if it returns the depot, then exit.
	     if( currIndex == startIndex ) stay=0;  

           }// end of if

         }// end of while

       }// end if 

       // if the edge list does not return to the starting base, then
       // we need to return this vehicle back to its originating depot.
       if( currIndex!=startIndex && 
	   //myGraph->doesEdgeExist(currIndex,startIndex) )
	   myGraph->canReachTo(currIndex,startIndex,currTime) )
       {
	  vrptime sstime = a.getStartDateSecondNode();
	  a = *myGraph->getEdgeFromTo(currIndex,startIndex);
	  a.setStartDate(sstime);
          currTime = a.getEndDate();
	  eList->addEdge(a);
	  currIndex = a.getEndNode();

       }// end of if	

       // EdgeList should be a cycle 
       if(currIndex==startIndex) 
       {
         //cout << *eList << endl << aCover << endl ;
         //cin >> enter;
         addVehicleDuty(eList);
       } else
	 { aCover.resetCover(eList); delete eList; }

     } else notExit = 0;
   }// end of while

}// end of patchUpRandom()

/*************************************************************************/
/*************************************************************************/
VRPSolution::solveRandom()
{ 
  // Define a cover 
  Cover aCover(myProblem->getNumberOfCustomers());

  cleanUpVehicles();

  patchUpRandom(aCover);

  return myCover->getNumberOfUncovered();

}// end of solveRandom()

/*************************************************************************/
/*************************************************************************/
void VRPSolution::patchUp(tSet<EdgePtr>* abandoned) 
{
  // need patchUp ?
  if( myCover->getNumberOfUncovered() ) patchUpRandom(*myCover,abandoned);

}// end of patchUp()

/*************************************************************************/
/*************************************************************************/
VRPSolution::doesContainEdge(int fr, int to) const {
  int ret=0;
  ListOfEdgeListsPtr tmp;
  for(int i=0;!ret&&i<getNumberOfVehicles();i++)
    if((tmp=getListOfEdgeListsIndex(i))!=NULL) ret=tmp->doesContainEdge(fr,to);
  return ret;
}// end of doesContainEdge()

/*************************************************************************/
/*************************************************************************/
#include "LocalSearch.h"

void VRPSolution::applyLocalSearch() 
{
   LocalSearch searcher;
   cout << " In VRPSolution::applyLocalSearch() " << endl; 
   //cout << *this << endl; 
   //char aaaa;
   //cin >> aaaa;
   VRPSolution* res = searcher.reorganizeRotations(this);
   //cout << " before copyFrom() " << endl; 
   if(res!=NULL&&res!=this) copyFrom(*res);
   cout << " Out VRPSolution::applyLocalSearch() " << endl; 

}// end of applyLocalSearch()

/*************************** T H E   E N D *******************************/

